import uuid
from typing import List, Type, Union, Callable

import riberry
from riberry import testing
from riberry.celery.background.events.events import process as _process_events


def setup_test_scenario(
        test_app: Type[testing.base.TestApplication],
        job_specification: testing.base.JobSpecification,
):
    """ Creates a new job for the given class from the given specification. """

    form: riberry.model.interface.Form = riberry.model.interface.Form.query().filter_by(
        internal_name=job_specification.form_internal_name,
    ).one()
    testing.util.run_until_successful(
        func=testing.waiters.wait_for_instance_online,
        kwargs=dict(instance_internal_name=form.instance.internal_name),
        time_limit=30,
        timeout_assertion_message='Application instance not online in time',
    )

    job = job_specification.create()

    executions = job.executions
    assert len(executions) == 1, f'Expected 1 execution, found {len(executions)}'
    test_app.execution_id = executions[0].id
    test_app.job_id = job.id
    test_app.job_values = {input_.internal_name: input_.value for input_ in job.values}
    test_app.job_files = {input_.internal_name: input_.binary for input_ in job.files}


def verify_execution_completion(
        execution: riberry.model.job.JobExecution,
        completion_status: str = 'SUCCESS',
        process_events: bool = True,
        time_limit: int = 30,
):
    """ Waits for a job to complete with the given status.

    The specified time limit is applied twice:
        1. Waiting for the execution to complete
        2. Waiting for the execution's events to complete processing
    """

    # Wait for the execution to complete
    testing.util.run_until_successful(
        func=testing.waiters.wait_for_execution_completion,
        kwargs=dict(execution=execution, expected_completion_status=completion_status),
        time_limit=time_limit,
        timeout_assertion_message='Execution did not complete in time',
    )

    # Manually process execution's events
    if process_events:
        _process_queued_events(execution=execution)

    # Wait for the execution's events to complete processing
    testing.util.run_until_successful(
        func=testing.waiters.wait_for_execution_events_to_process,
        args=(execution,),
        time_limit=time_limit,
        timeout_assertion_message='Execution events did not complete in time',
    )


def verify_streams(execution: riberry.model.job.JobExecution):
    """ Ensures that all streams and steps are successful and "correct". """

    for stream in execution.streams:
        assert all((stream.created, stream.started, stream.completed, stream.updated)), 'Not all stream dates populated'
        assert stream.completed >= stream.started, 'Stream dates are not correct'
        assert stream.status == 'SUCCESS'
        for step in stream.steps:
            assert all((step.created, step.started, step.completed, step.updated)), 'Not all step dates populated'
            assert step.completed >= step.started, 'Step dates are not correct'


def create_job(
        form_internal_name: str,
        username: str,
        job_prefix: str = 'TEST_',
        job_name: Union[Callable[[], str], str] = uuid.uuid4,
        input_values: dict = None,
        input_files: dict = None,
        execute: bool = True,
):
    """ Creates a test job for the given form. """
    user = riberry.model.auth.User.query().filter_by(username=username).one()
    with riberry.services.policy.policy_scope(user=user):
        job = riberry.services.job.create_job(
            riberry.services.form.form_by_internal_name(internal_name=form_internal_name).id,
            name=f'{job_prefix}{job_name() if callable(job_name) else job_name}',
            input_values=input_values or {},
            input_files=input_files or {},
            execute=execute,
        )
        riberry.model.conn.commit()
        return job


def delete_jobs(form_internal_name: str, job_prefix: str = 'TEST_'):
    """ Deletes a job for the given form with the given prefix. """

    form_id = riberry.services.form.form_by_internal_name(internal_name=form_internal_name).id
    jobs: List[riberry.model.job.Job] = riberry.model.job.Job.query().filter(
        (riberry.model.job.Job.form_id == form_id) & riberry.model.job.Job.name.startswith(job_prefix)
    ).all()

    for job in jobs:
        riberry.model.conn.delete(instance=job)

    riberry.model.conn.commit()


def _process_queued_events(execution: riberry.model.job.JobExecution):
    """ Process the queued events for the given execution. """
    root_id = execution.task_id
    _process_events(query_extension=lambda q: q.filter_by(root_id=root_id))



import os
import signal
import threading
from typing import AnyStr, Dict, Optional

import riberry
from riberry.app import RiberryApplication, current_context as cxt
from . import tasks
from .task_queue import TaskQueue, Task, TaskDefinition
from .tracker import PoolExecutionTracker

log = riberry.log.make(__name__)


def entry_point(form_name, **kwargs):
    form: riberry.model.interface.Form = riberry.model.interface.Form.query().filter_by(internal_name=form_name).one()
    try:
        app: RiberryApplication = RiberryApplication.by_name(form.application.internal_name)
    except KeyError:
        RiberryApplication(name=form.application.internal_name, backend=RiberryPoolBackend())
        app: RiberryApplication = RiberryApplication.by_name(form.application.internal_name)

    return app.entry_point(form_name, **kwargs)


class RiberryPoolBackend(riberry.app.backends.RiberryApplicationBackend):
    _local = threading.local()

    execution_tracker: PoolExecutionTracker

    def __init__(self):
        super().__init__(instance=None)
        self.task_queue: TaskQueue = TaskQueue(backend=self, limit=3)
        self.tasks = {}
        self._exit = threading.Event()
        self._threads = []
        self._pool_execution_tracker = PoolExecutionTracker(backend=self)

    def _create_thread(self, thread_name, target):
        thread = threading.Thread(name=thread_name, target=target)
        self._threads.append(thread)
        return thread

    def start(self):
        log.debug('Starting application %s', riberry.app.current_riberry_app.name)

        for sig in ('SIGTERM', 'SIGINT', 'SIGHUP'):
            if hasattr(signal, sig):
                signal.signal(getattr(signal, sig), self._stop_signal)

        for thread in self._threads:
            thread.start()

        log.info('Started application %s', riberry.app.current_riberry_app.name)
        for thread in self._threads:
            thread.join()

    # noinspection PyUnusedLocal
    def _stop_signal(self, signum, frame):
        self.stop()

    def stop(self):
        log.info('Stopping application %s', riberry.app.current_riberry_app.name)
        self._exit.set()

    def initialize(self):
        self._create_thread('backend.executor', lambda: tasks.run_task(
            name='Task Executor',
            func=lambda: tasks.execution_listener(self.task_queue),
            interval=0,
            exit_event=self._exit,
        ))

        self._create_thread('backend.queue_external', lambda: tasks.run_task(
            name='External Task Receiver',
            func=lambda: tasks.queue_receiver_tasks(self.task_queue),
            interval=5,
            exit_event=self._exit,
        ))

        self._create_thread('backend.background', lambda: tasks.run_task(
            name='Background Operations',
            func=lambda: tasks.background(self.task_queue),
            interval=5,
            exit_event=self._exit,
        ))

    def default_addons(self) -> Dict[AnyStr, 'riberry.app.addons.Addon']:
        return {}

    def start_execution(self, execution_id, root_id, entry_point) -> AnyStr:
        self.task_queue.submit_entry_task(
            execution_id=execution_id,
            root_id=root_id,
            entry_point=entry_point
        )
        return root_id

    def create_receiver_task(self, external_task_id, validator):
        cxt.data.set(f'external:{external_task_id}:validator', validator)

    def register_task(self, func, name=None, stream=None, step=None, **options):
        name = name or riberry.app.util.misc.function_path(func)
        assert name not in self.tasks, f'Multiple registrations for task {name!r}'
        assert 'after' in options and not self.external_task_callback(options['after']), (
            f'"after" argument not supplied to task {name!r}'
        )
        self.tasks[name] = TaskDefinition(
            func=func,
            name=name,
            stream=stream or self.default_stream_name,
            step=step or name,
            options=options,
        )

    def task_by_name(self, name: AnyStr) -> TaskDefinition:
        return self.tasks[name]

    def external_task_callback(self, name: AnyStr) -> Optional[TaskDefinition]:
        for task in self.tasks.values():
            if task.options.get('after') == name:
                return task
        return None

    def active_task(self):
        return getattr(self._local, 'task', None)

    def set_active_task(self, task: Task):
        self._local.task = task

    def _execution_tracker(self):
        return self._pool_execution_tracker



import click

from riberry.util import groups
from .base import admin


@click.group(help='Collection of group function')
def group():
    pass


@group.command('add', help='Add a user to a group')
@click.option('--username', '-u', prompt='Username', help="User's username")
@click.option('--group', '-g', prompt='Group', help="Group's name")
def add(username, group):
    try:
        groups.add_user_to_group(username=username, group_name=group)
        print(f'Added {username} to group {group}')
    except Exception as exc:
        print(str(exc))
        exit(1)


@group.command('remove', help='Remove a user from a group')
@click.option('--username', '-u', prompt='Username', help="User's username")
@click.option('--group', '-g', prompt='Group', help="Group's name")
def remove(username, group):
    try:
        groups.remove_user_from_group(username=username, group_name=group)
        print(f'Removed {username} from group {group}')
    except Exception as exc:
        print(str(exc))
        exit(1)


admin.add_command(group)



import os
from typing import Union, Any
import sqlalchemy
import sqlalchemy.orm
import sqlalchemy.pool
import sqlalchemy.engine

from . import misc, application, group, auth, interface, job, base
from ..util.misc import import_from_string
from .. import log

log = log.make(__name__)

ScopedSessionExt = Union[sqlalchemy.orm.Session, sqlalchemy.orm.scoping.ScopedSession]


class __ModelProxy:
    raw_session: ScopedSessionExt = None
    raw_engine: sqlalchemy.engine.Engine = None
    _pid_created_in: int = os.getpid()

    def _check_pid(self):
        """ Disposes SQLAlchemy engine if current process is a fork. """

        current_pid = os.getpid()
        if self._pid_created_in != current_pid:
            log.debug(f'riberry.model.conn:: Process forked from pid {self._pid_created_in}, disposing of sqla engine')
            self._pid_created_in = current_pid
            self.dispose_engine()

    def __getattr__(self, item: str) -> Any:
        self._check_pid()
        return getattr(self.raw_session, item)

    def __enter__(self) -> '__ModelProxy':
        self._check_pid()
        self.raw_session()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            self.raw_session.rollback()
        self.dispose_session()

    def dispose_engine(self):
        """ Disposes of the current SQLAlchemy engine. """
        self.dispose_session()
        try:
            self.raw_engine.dispose()
        except:
            log.exception('riberry.model.conn:: Encountered an error while disposing current sqla engine')

    def dispose_session(self):
        """ Disposes of the current SQLAlchemy session. """
        try:
            self.raw_session.remove()
        except:
            log.exception('riberry.model.conn:: Encountered an error while removing current sqla session')


conn: Union[ScopedSessionExt, __ModelProxy] = __ModelProxy()


def init(url='sqlite://', engine_settings=None, connection_arguments=None):
    engine_defaults = dict(
        poolclass='sqlalchemy.pool:QueuePool',
        pool_use_lifo=True,
        pool_pre_ping=True,
        pool_recycle=360,
    )
    engine_settings = {**engine_defaults, **(engine_settings or {})}
    engine_settings['poolclass'] = import_from_string(engine_settings.pop('poolclass') or 'sqlalchemy.pool:QueuePool')
    connection_arguments = connection_arguments or {}

    __ModelProxy.raw_engine = sqlalchemy.create_engine(
        url,
        **engine_settings,
        connect_args=connection_arguments,
    )
    __ModelProxy.raw_session = sqlalchemy.orm.scoped_session(sqlalchemy.orm.sessionmaker(bind=__ModelProxy.raw_engine))
    base.Base.metadata.create_all(__ModelProxy.raw_engine)



from collections import defaultdict
from typing import Callable, Tuple, Dict, Any, Iterable

import riberry

ArtifactProcessor = Callable[[riberry.model.job.JobExecutionArtifact], Tuple[str, dict]]


def _compare(expected: Any, actual: Any, failure_message: str):
    """ Simples equality assertion function.

    Useful when testing with pytest which will log the input parameters
    in case of a failure.
    """
    assert expected == actual, failure_message


def compare_streams(
        execution: riberry.model.job.JobExecution,
        expected: Dict[str, Dict[str, int]]
):
    """ Compares the given streams with the execution's actual streams.

     Example of `expected` parameter:

     {
        "STREAM_NAME_1": {     # stream name
            "STEP_NAME_1": 1,  # key: step name, value: step count,
            "STEP_NAME_2": 2   # key: step name, value: step count,
        },
        "STREAM_NAME_2": {     # stream name
            "STEP_NAME_1": 1   # key: step name, value: step count,
        }
    }
    """

    actual = get_actual_stream_step_counts(execution)
    _compare(expected, actual, 'Streams do not match')


def compare_artifacts(
        execution: riberry.model.job.JobExecution,
        expected: dict,
        artifact_processor: ArtifactProcessor = None
):
    """ Compares the given streams with the execution's actual streams.

    Example of default `expected` parameter:

    {
        "ARTIFACT_1_FILENAME": {
            "ARTIFACT_NAME": ...,
            "ARTIFACT_TYPE": ...,
            "ARTIFACT_CATEGORY": ...
        },
        "ARTIFACT_2_FILENAME": {
            ...
        }
    }
    """

    artifact_processor = artifact_processor or default_artifact_processor
    actual = {key: value for key, value in get_actual_artifacts(execution, artifact_processor)}
    _compare(expected, actual, 'Artifacts do not match')


def get_actual_stream_step_counts(execution: riberry.model.job.JobExecution) -> Dict[str, Dict[str, int]]:
    """ Returns all streams and the count tallies for steps under those streams. """

    stream_counts_actual = defaultdict(lambda: defaultdict(int))
    for stream in execution.streams:
        for step in stream.steps:
            stream_counts_actual[stream.name][step.name] += 1

    return {
        stream_name: dict(steps)
        for stream_name, steps in stream_counts_actual.items()
    }


def default_artifact_processor(artifact: riberry.model.job.JobExecutionArtifact) -> Tuple[str, dict]:
    """ Returns basic information for generated artifacts """

    return artifact.filename, {
        'name': artifact.name,
        'type': artifact.type.value,
        'category': artifact.category,
    }


def get_actual_artifacts(
        execution: riberry.model.job.JobExecution,
        artifact_processor: ArtifactProcessor
) -> Iterable[Tuple[str, dict]]:
    """ Returns an iterable of all artifacts for the given execution. """

    return (artifact_processor(artifact) for artifact in execution.artifacts)



from typing import List

from sqlalchemy import desc, asc

import riberry

log = riberry.log.make(__name__)


class RiberryExecutionTracker:

    def __init__(self, backend):
        self.backend: riberry.app.backends.RiberryApplicationBackend = backend

    def check_stale_executions(self, app_instance: riberry.model.application.ApplicationInstance):
        executions: List[riberry.model.job.JobExecution] = riberry.model.job.JobExecution.query().filter(
            riberry.model.job.JobExecution.status.in_(('ACTIVE', 'READY'))
        ).join(riberry.model.job.Job).order_by(
            desc(riberry.model.job.JobExecution.priority),
            asc(riberry.model.job.JobExecution.created)
        ).filter_by(instance=app_instance).all()

        if not executions:
            return

        self._check_stale_executions(executions=executions, app_instance=app_instance)

    def _check_stale_executions(
            self,
            executions: List[riberry.model.job.JobExecution],
            app_instance: riberry.model.application.ApplicationInstance
    ):
        raise NotImplementedError

    def track_execution(self, root_id: str, app_instance: riberry.model.application.ApplicationInstance):
        raise NotImplementedError

    def _artifact_message(self, execution: riberry.model.job.JobExecution):
        raise NotImplementedError

    def _cancel_execution(self, execution: riberry.model.job.JobExecution):
        log.warning(
            f'Cancelling stale execution: id={execution.id}, root={execution.task_id!r}, job={execution.job.name!r}')

        riberry.app.actions.artifacts.create_artifact(
            name='Workflow Cancelled',
            type=riberry.model.job.ArtifactType.error,
            category='Fatal',
            filename='fatal.log',
            content=self._artifact_message(execution),
            task_id=execution.task_id,
            root_id=execution.task_id,
        )

        riberry.app.actions.executions.execution_complete(
            task_id=execution.task_id,
            root_id=execution.task_id,
            status='FAILURE',
            stream=None,
            context=riberry.app.current_context,
        )



from typing import Callable, Optional, List

import pendulum
from sqlalchemy import desc, asc

import riberry
from riberry.app import current_riberry_app
from . import actions, env

log = riberry.log.make(__name__)


def echo():
    with riberry.model.conn:
        app_instance = env.get_instance_model()

        heartbeat = riberry.model.application.Heartbeat.query().filter_by(instance=app_instance).first()
        if not heartbeat:
            heartbeat = riberry.model.application.Heartbeat(instance=app_instance)
            riberry.model.conn.add(heartbeat)

        heartbeat.updated = pendulum.DateTime.utcnow()
        riberry.model.conn.commit()


def poll(
        track_executions: bool = True,
        filter_func: Optional[Callable[[riberry.model.job.JobExecution], bool]] = None,
):
    with riberry.model.conn:
        app_instance = env.get_instance_model()

        if track_executions:
            current_riberry_app.backend.execution_tracker.check_stale_executions(app_instance=app_instance)

        instance_name = app_instance.internal_name
        if app_instance.status != 'online':
            log.debug(f'Instance {instance_name!r} is not online, skipped polling executions')
            return

        if app_instance.active_schedule_value('accept', default='Y') == 'N':
            log.debug(f'Instance {instance_name!r} is not accepting new executions, skipped polling executions')
            return

        executions: List[riberry.model.job.JobExecution] = riberry.model.job.JobExecution.query().filter(
            riberry.model.job.JobExecution.status == 'RECEIVED'
        ).join(riberry.model.job.Job).order_by(
            desc(riberry.model.job.JobExecution.priority),
            asc(riberry.model.job.JobExecution.created),
            asc(riberry.model.job.JobExecution.id),
        ).filter_by(instance=app_instance).all()

        for execution in executions:
            if execution_limit_reached(app_instance=app_instance):
                log.debug(f'Instance {instance_name!r} has reached the allowed limit of active/ready executions')
                return
            if callable(filter_func) and not filter_func(execution):
                continue
            execution_task_id = actions.executions.queue_job_execution(
                execution=execution, track_executions=track_executions)
            log.info(f'Queueing execution: id={execution.id!r}, root={execution_task_id!r}, job={execution.job.name!r}')


def execution_limit_reached(app_instance: riberry.model.application.ApplicationInstance) -> bool:
    limit_raw = str(app_instance.active_schedule_value('limit'))
    if not limit_raw.isdigit():
        return False

    limit = int(limit_raw)
    if limit > 0:
        active_execution_count = riberry.model.job.JobExecution.query().filter(
            riberry.model.job.JobExecution.status.in_(('READY', 'ACTIVE'))
        ).join(
            riberry.model.job.Job
        ).filter_by(
            instance=app_instance
        ).count()
        return active_execution_count >= limit

    return False


def refresh():
    with riberry.model.conn:
        app_instance = env.get_instance_model()
        actions.shared_data.update_all_data_items(app_instance=app_instance)
        actions.reports.update_all_reports(app_instance=app_instance)



from . import (
    base,           # Classes used to create test suites and scenarios
    util,           # Re-usable utilities
    waiters,        # Functions intended to be called with util.run_until_successful
    comparators,    # Execution-related comparison functions (streams, steps, artifacts)
    helpers,        # Helper functions used to setup, configure and verify test scenarios
)



from celery import signals

import riberry


@signals.celeryd_after_setup.connect
def celeryd_after_setup(**_):
    """
    Dispose of Riberry's SQLAlchemy engine after the Celery daemon has completed setup.
    """
    riberry.model.conn.dispose_engine()


@signals.worker_process_init.connect
def worker_process_init(*args, **kwargs):
    """
    Dispose of Riberry's SQLAlchemy engine in newly initialized worker processes.

    This is used to prevent database sessions being mistakenly used by both
    the parent and child process.
    """

    riberry.model.conn.dispose_engine()


@signals.worker_ready.connect
def worker_ready(sender, **_):
    """
    Patch "celery.concurrency.asynpool.AsynPool.on_process_up" to dispose of
    Riberry's SQLAlchemy engine whenever a new worker process is created via
    prefork.

    This is used to prevent database sessions being mistakenly used by both
    the parent and child process.
    """

    pool = getattr(sender.pool, '_pool', None)
    if pool is None:
        return

    on_process_up_original = pool.on_process_up

    def on_process_up(proc):
        riberry.model.conn.dispose_engine()
        on_process_up_original(proc)

    pool.on_process_up = on_process_up
    riberry.model.conn.dispose_engine()


@signals.before_task_publish.connect
def before_task_publish(sender, headers, body, **_):
    """ Inform Riberry of newly created tasks. """
    
    _, task_input, *_ = body

    riberry.app.util.task_transitions.task_created(
        context=riberry.app.current_context,
        task_id=headers['id'],
        stream=task_input.get('__rib_stream'),
        step=task_input.get('__rib_step'),
        props=dict(
            stream_start=task_input.get('__rib_stream_start') or False,
        ),
    )



import click

from riberry.util.__main__ import add_user
from .base import admin


@click.group(help='Collection of user function')
def user():
    pass


user.add_command(add_user, name='create')
admin.add_command(user)



from typing import List

import riberry

log = riberry.log.make(__name__)


class PoolExecutionTracker(riberry.app.backends.RiberryExecutionTracker):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tracked_executions = set()

    def _check_stale_executions(
            self,
            executions: List[riberry.model.job.JobExecution],
            app_instance: riberry.model.application.ApplicationInstance
    ):
        for execution in executions:
            if execution.task_id not in self.tracked_executions and not self.has_pending_tasks(execution=execution):
                self._cancel_execution(execution=execution)

    def track_execution(self, root_id: str, app_instance: riberry.model.application.ApplicationInstance):
        if root_id not in self.tracked_executions:
            log.debug(f'Tracking execution: root={root_id!r}')
            self.tracked_executions.add(root_id)

    def _artifact_message(self, execution: riberry.model.job.JobExecution):
        return (
            f'The current executions\'s ID ({execution.task_id}) is no longer being tracked and has '
            f'therefore been cancelled. This usually occurs when the worker has terminated while a '
            f'task was either internally queued for execution or actively executing.'
        )

    @staticmethod
    def has_pending_tasks(execution: riberry.model.job.JobExecution):
        for task in execution.external_tasks:
            if task.status != 'COMPLETE':
                return True
        return False



import uuid
from typing import Union, Callable

import riberry
from riberry import testing


class JobSpecification:
    """ Specification for a job to be created. """

    def __init__(
            self,
            form_internal_name: str,
            username: str,
            job_prefix: str = 'TEST_',
            job_name: Union[Callable[[], str], str] = uuid.uuid4,
            input_values: dict = None,
            input_files: dict = None,
            execute: bool = True,
    ):
        self.form_internal_name = form_internal_name
        self.username = username
        self.input_values = input_values or {}
        self.input_files = input_files or {}
        self.job_prefix = job_prefix
        self.job_name = job_name
        self.execute = execute

    def create(self) -> riberry.model.job.Job:
        return testing.helpers.create_job(
            form_internal_name=self.form_internal_name,
            username=self.username,
            job_prefix=self.job_prefix,
            job_name=self.job_name,
            input_values=self.input_values,
            input_files=self.input_files,
            execute=self.execute,
        )


class TestApplication:
    """ Helper class used to test job executions.

    Call `setup_test_scenario` when setting up the class to populate the
    default class values.
    """

    job_id: int = None
    execution_id: int = None
    job_values: dict = None
    job_files: dict = None

    @classmethod
    def __job(cls):
        return riberry.model.job.Job.query().filter_by(id=cls.job_id).one()

    @classmethod
    def __execution(cls) -> riberry.model.job.JobExecution:
        return riberry.model.job.JobExecution.query().filter_by(id=cls.execution_id).one()

    @classmethod
    def setup_test_scenario(cls, job_specification: JobSpecification):
        testing.helpers.setup_test_scenario(test_app=cls, job_specification=job_specification)

    @classmethod
    def delete_test_job(cls):
        riberry.model.conn.delete(cls.__job())
        riberry.model.conn.commit()

    @property
    def job(self) -> riberry.model.job.Job:
        return self.__job()

    @property
    def execution(self) -> riberry.model.job.JobExecution:
        return self.__execution()



from threading import Event

import riberry
from .background import background
from .executor import execution_listener
from .external_task_receiver import queue_receiver_tasks

log = riberry.log.make(__name__)


def run_task(name, func, interval, exit_event: Event):
    log.debug('Started task %s', name)

    while not exit_event.is_set():
        try:
            func()
        except:
            log.exception('Error occurred while processing task %s', name)
        finally:
            if interval:
                exit_event.wait(interval)

    log.debug('Stopped task %s', name)



import importlib

import click

from .base import run


@run.command('web', help='Start Riberry\'s web application')
@click.option('--module', '-m', show_default=True, default='riberry_web:main', help='Callable to start the web server.')
@click.option('--host', '-h', show_default=True, default='127.0.0.1', help='Bind to given host.')
@click.option('--port', '-p', show_default=True, default=5445, help='Bind to given port.')
@click.option('--log-level', '-l', show_default=True, default='info', help='Logging level.')
@click.pass_context
def run_webapp(cxt, module: str, host: str, port: int, log_level: str):
    module_dot_path, callable_name = module.split(':')
    try:
        mod = importlib.import_module(module_dot_path)
    except ModuleNotFoundError:
        print(f'Could not find module {module_dot_path}.')
        return cxt.exit(1)

    try:
        callable_obj = getattr(mod, callable_name)
    except AttributeError:
        print(f'Could not find callable {callable_name} in module {module_dot_path}.')
        return cxt.exit(1)

    callable_obj(host=host, port=port, log_level=log_level)



from .base import AddonStartStopStep
import riberry


class BackgroundTasks(riberry.app.addons.Addon):

    def register(self, riberry_app: 'riberry.app.base.RiberryApplication'):
        class ConcreteBackgroundTasksStep(BackgroundTasksStep):
            rib = riberry_app

        riberry_app.backend.steps['worker'].add(ConcreteBackgroundTasksStep)


class BackgroundTasksStep(AddonStartStopStep):
    requires = {'celery.worker.components:Timer'}

    def __init__(self, worker, **_):
        super().__init__(worker=worker, interval=1.0)
        self.lock = riberry.app.util.redis_lock.RedisLock(
            name='step:background',
            on_acquired=self.on_lock_acquired,
            interval=900,
        )

    @staticmethod
    def on_lock_acquired():
        riberry.app.tasks.echo()
        riberry.app.tasks.poll()
        riberry.app.tasks.refresh()

    def should_run(self) -> bool:
        return True

    def run(self):
        redis_instance = riberry.celery.util.celery_redis_instance()
        self.lock.run(redis_instance=redis_instance)



from typing import Callable, Iterable, Optional

import time


def run_until_successful(
        func: Callable,
        args: Optional[Iterable] = None,
        kwargs: Optional[dict] = None,
        time_limit: int = 10,
        timeout_assertion_message: Optional[str] = None,
        wait_interval: int = 1,
):
    """ Runs the given function at the given interval until it returns True or exceeds the time limit. """

    start_time = time.time()

    while True:
        if func(*args or (), **kwargs or {}):
            return
        assert time.time() - start_time < time_limit, timeout_assertion_message or (
                timeout_assertion_message or f'Function {func.__name__} failed'
        )
        time.sleep(wait_interval)



import importlib


def import_from_string(module_path: str):
    if ':' in module_path:
        module_path, object_name = module_path.split(':')
    else:
        object_name = None

    try:
        mod = importlib.import_module(module_path)
    except ModuleNotFoundError:
        print(f'Could not find module {module_path}.')

    return getattr(mod, object_name) if object_name else mod



import logging

from riberry import log

__version__ = '0.10.22'

log.root_name = __name__
log.logger = logging.getLogger(log.root_name)
log.init()

from riberry import config, plugins, model, celery, policy, services, exc, app

config.config.enable()
config.config.authentication.enable()



import json

import click

from riberry.util import config_importer
from .base import admin


@click.group(help='Collection of model function')
def model():
    pass


@model.command('import', help='Imports and synchronises Riberry\'s models from config')
@click.option('--config-path', '-p', prompt='Configuration path', help='YAML file containing riberry model.')
@click.option('--commit/--rollback', '-c/-r', prompt='Dry run', is_flag=True,
              help='Gathers all database changes without commits.', default=False)
@click.option('--app', '-a', 'apps', help='Restricts the import to the specified app', multiple=True)
def importer(config_path, commit, apps):
    changes = config_importer.import_from_file(config_path, dry_run=not commit, restrict_apps=apps)
    print(json.dumps(changes, indent=2))


admin.add_command(model)



import threading
import uuid
from functools import wraps
from queue import Queue, Full

import riberry


class TaskDefinition:

    def __init__(self, func, name, stream, step, options):
        self.func = func
        self.name = name
        self.stream = stream
        self.step = step
        self.options = options


class Task:

    def __init__(self, task_id: str, execution_id: int, definition: TaskDefinition):
        self.id = task_id
        self.execution_id = execution_id
        self.definition = definition


class TaskCounter:

    def __init__(self):
        self._value = 0
        self._lock = threading.RLock()

    def increment(self):
        with self._lock:
            self._value += 1
            return self._value

    def decrement(self):
        with self._lock:
            self._value -= 1
            return self._value

    @property
    def lock(self):
        return self._lock

    @property
    def value(self):
        return self._value


class TaskQueue:
    queue_cls = Queue

    def __init__(self, backend, queue=None, limit=None):
        self.backend: riberry.app.backends.impl.pool.RiberryPoolBackend = backend
        self.queue = queue or self.queue_cls()
        self.limit = limit
        self.counter = TaskCounter()

    @property
    def lock(self):
        return self.counter.lock

    def limit_reached(self):
        return bool(self.limit is not None and self.counter.value >= self.limit)

    def submit_receiver_task(self, external_task: riberry.model.job.JobExecutionExternalTask):
        self.backend.execution_tracker.track_execution(
            root_id=external_task.job_execution.task_id,
            app_instance=external_task.job_execution.job.instance,
        )
        self._submit(make_receiver_task(backend=self.backend, external_task=external_task))

    def submit_entry_task(self, execution_id: int, root_id: str, entry_point: riberry.app.base.EntryPoint):
        self._submit(make_entry_task(execution_id=execution_id, root_id=root_id, entry_point=entry_point))

    def _submit(self, task: Task):
        with self.lock:
            if self.limit_reached():
                raise Full
            self.queue.put_nowait(task)
            self.counter.increment()


def _make_external_task_wrapper(task_id, func):
    @wraps(func)
    def wrapper():
        task = riberry.model.job.JobExecutionExternalTask.query().filter_by(id=task_id).one()
        task.status = 'COMPLETE'
        riberry.model.conn.commit()
        func(task.output_data)

    return wrapper


def make_receiver_task(backend, external_task: riberry.model.job.JobExecutionExternalTask) -> Task:
    definition = backend.external_task_callback(external_task.name)
    return Task(
        task_id=str(uuid.uuid4()),
        execution_id=external_task.job_execution.id,
        definition=TaskDefinition(
            func=_make_external_task_wrapper(external_task.id, definition.func),
            name=definition.step,
            stream=definition.stream,
            step=definition.step,
            options=definition.options,
        ),
    )


def make_entry_task(execution_id: int, root_id: str, entry_point: riberry.app.base.EntryPoint):
    return Task(
        task_id=root_id,
        execution_id=execution_id,
        definition=TaskDefinition(
            func=entry_point.func,
            name=entry_point.step,
            stream=entry_point.stream,
            step=entry_point.step,
            options={},
        )
    )



import riberry


def wait_for_instance_online(instance_internal_name: str) -> bool:
    """ Returns true only when the given instance's status is "online". """
    with riberry.model.conn:
        instance: riberry.model.application.ApplicationInstance = riberry.model.application.ApplicationInstance.query().filter_by(
            internal_name=instance_internal_name,
        ).one()
        return instance.status == 'online'


def wait_for_execution_completion(execution: riberry.model.job.JobExecution, expected_completion_status: str) -> bool:
    """ Returns true only when the given execution's status matches the expected completion status.

    If the expected completion status doesn't match, an assertion error will be raised.
    """
    riberry.model.conn.expire(execution)
    if execution.status == expected_completion_status:
        return True

    assert not (
            (execution.status == 'SUCCESS' and expected_completion_status == 'FAILURE') or
            (execution.status == 'FAILURE' and expected_completion_status == 'SUCCESS')
    ), f'Expected state {expected_completion_status}, got {execution.status}'

    return False


def wait_for_execution_events_to_process(execution: riberry.model.job.JobExecution) -> bool:
    """ Returns true only when the given execution has no background event queued up for processing. """

    riberry.model.conn.expire(execution)
    return riberry.model.misc.Event.query().filter_by(root_id=execution.task_id).count() == 0



import atexit
import subprocess

import click

from riberry import log
from .base import run

log = log.make(__name__)


def _kill(processes):
    for process in processes:
        try:
            if process.poll() is None:
                process.kill()
        except Exception:
            log.exception(f'Failed to kill process {process}')


@run.command(help='Start Riberry\'s core background celery app')
@click.option('--log-level', '-l', default='ERROR', help='Log level')
def core(log_level):
    process_beat = subprocess.Popen([
        'celery', 'beat',
        '-A', 'riberry.celery.background',
        '-l', log_level.lower(),
    ])

    process_background = subprocess.Popen([
        'celery', 'worker',
        '-A', 'riberry.celery.background',
        '-l', log_level.lower(),
        '-c', '1',
        '-Q', 'riberry.background.custom,'
              'riberry.background.schedules,'
              'riberry.background.events,'
              'riberry.background.metrics',
    ])

    processes = process_background, process_beat

    atexit.register(lambda: _kill(processes=processes))

    try:
        process_background.wait()
        process_beat.wait()
    except KeyboardInterrupt:
        log.info('Exiting riberry core...')



import importlib
import os

import click

import riberry
from riberry.app.backends.impl.pool.base import RiberryPoolBackend
from riberry.app.backends.impl.pool.log import configure as log_configure
from .base import run


@run.command(help='Start custom Riberry pool application')
@click.option('--module', '-m', required=True, help='Module containing Riberry pool application')
@click.option('--instance', '-i', help='Riberry application instance to run')
@click.option('--log-level', '-l', default='ERROR', help='Log level')
@click.option('--concurrency', '-c', default=None, help='Task concurrency', type=int)
def pool(module, instance, log_level, concurrency):
    if instance is not None:
        os.environ['RIBERRY_INSTANCE'] = instance

    log_configure(log_level=log_level)
    importlib.import_module(module)
    backend: RiberryPoolBackend = riberry.app.current_riberry_app.backend
    backend.task_queue.limit = concurrency
    backend.start()



from typing import List

import riberry
from riberry.app import current_context as ctx
from ..task_queue import TaskQueue


def ready_external_tasks() -> List[riberry.model.job.JobExecutionExternalTask]:
    return riberry.model.conn.query(
        riberry.model.job.JobExecutionExternalTask
    ).filter_by(
        status='READY',
    ).join(riberry.model.job.JobExecution).filter_by(
        status='ACTIVE',
    ).join(riberry.model.job.Job).filter_by(
        instance=ctx.current.riberry_app_instance,
    ).all()


def queue_receiver_tasks(queue: TaskQueue):
    with queue.lock:
        if not queue.limit_reached():
            with riberry.model.conn:
                external_tasks = ready_external_tasks()
                while external_tasks and not queue.limit_reached():
                    queue.submit_receiver_task(external_tasks.pop())



from typing import Optional

import riberry
from ..util.events import create_event

__stream_cache = {}


def try_cache_stream(root_id: str, stream_name: str, stream_state: str):
    key = root_id, stream_name, stream_state
    if key not in __stream_cache:
        __stream_cache[key] = None
        if len(__stream_cache) > 10000:
            evict_key = next(iter(__stream_cache))
            __stream_cache.pop(evict_key)
        return True
    return False


def task_created(context, task_id: str, stream: Optional[str], step: Optional[str], props: dict):
    try:
        root_id = context.current.root_id
    except:
        return

    if props.get('stream_start'):
        if try_cache_stream(root_id=root_id, stream_name=stream, stream_state='QUEUED'):
            create_event(
                name='stream',
                root_id=root_id,
                task_id=task_id,
                data={
                    'stream': stream,
                    'state': 'QUEUED',
                }
            )

    if step and riberry.app.current_riberry_app.config.enable_steps:
        create_event(
            name='step',
            root_id=root_id,
            task_id=task_id,
            data={
                'stream': str(stream),
                'step': str(step),
                'state': 'QUEUED',
            }
        )


def task_active(context, props: dict):
    task_id = context.current.task_id
    root_id = context.current.root_id

    stream = context.current.stream
    step = context.current.step

    if not stream:
        return

    if 'stream_start' in props:
        if try_cache_stream(root_id=root_id, stream_name=stream, stream_state='ACTIVE'):
            create_event(
                name='stream',
                root_id=root_id,
                task_id=task_id,
                data={
                    'stream': str(stream),
                    'state': 'ACTIVE',
                }
            )

    if step and riberry.app.current_riberry_app.config.enable_steps:
        create_event(
            name='step',
            root_id=root_id,
            task_id=task_id,
            data={
                'stream': str(stream),
                'step': str(step),
                'state': 'ACTIVE',
            }
        )


def task_complete(context, props: dict, state: str):
    task_id = context.current.task_id
    root_id = context.current.root_id

    stream = context.current.stream
    step = context.current.step

    if not stream:
        return

    if props.get('stream_start') and state in ('RETRY', 'FAILURE'):
        create_event(
            name='stream',
            root_id=root_id,
            task_id=task_id,
            data={
                'stream': str(stream),
                'state': state,
            }
        )

    if props.get('stream_end'):
        create_event(
            name='stream',
            root_id=root_id,
            task_id=task_id,
            data={
                'stream': str(stream),
                'state': 'SUCCESS' if state == 'IGNORED' else state or 'FAILURE',
            }
        )

    if step and riberry.app.current_riberry_app.config.enable_steps:
        create_event(
            name='step',
            root_id=root_id,
            task_id=task_id,
            data={
                'stream': str(stream),
                'step': str(step),
                'state': 'SUCCESS' if state == 'IGNORED' else state or 'FAILURE',
            }
        )



import abc
from threading import local


class PolicyContextStore(abc.ABC):

    def get(self, item, default=None):
        raise NotImplementedError

    def set(self, item, value):
        raise NotImplementedError


class ThreadLocalPolicyContextStore(PolicyContextStore):
    _local = local()

    def get(self, item, default=None):
        return getattr(self._local, item, default)

    def set(self, item, value):
        setattr(self._local, item, value)



from collections import defaultdict, namedtuple
from typing import List, DefaultDict

import pendulum
from sqlalchemy import func
from sqlalchemy.orm import Query

from celery.utils.log import logger
import riberry
from riberry.model.job import JobExecutionMetric, JobExecutionStreamStep, JobExecutionStream, Job, JobExecution

MetricKey = namedtuple('MetricKey', [
    'epoch_start',
    'epoch_end',
    'job_execution_id',
    'stream_name',
    'step_name',
])


def round_down(num: int, interval: int) -> int:
    """ Rounds down the given number to the nearest multiple of the given interval. """
    return int(num - (num % interval))


def latest_query_steps(epoch_start: int, limit: int = None) -> Query:
    """ Returns a query instance to retrieve the latest steps to process. """

    start_datetime = pendulum.from_timestamp(epoch_start, tz=pendulum.UTC)
    query: Query = riberry.model.conn.query(
        JobExecutionStreamStep.name,
        JobExecutionStreamStep.started,
        JobExecutionStreamStep.completed,
        JobExecutionStream.name,
        JobExecutionStream.job_execution_id,
        Job.form_id,
    ).filter(
        JobExecutionStreamStep.completed >= start_datetime
    ).order_by(JobExecutionStreamStep.completed.asc()).join(
        JobExecutionStream,
    ).join(
        JobExecution,
    ).join(
        Job,
    )

    if limit:
        query = query.limit(limit)

    return query


def retrieve_latest_epoch_start() -> int:
    """ Retrieves the max processed epoch start value. """
    return riberry.model.conn.query(func.max(JobExecutionMetric.epoch_start)).first()[0] or 0


def updated_metrics(time_interval: int, limit: int) -> List[JobExecutionMetric]:
    """ Updates/creates and returns metrics for the latest created steps. """

    updated: List[JobExecutionMetric] = []
    mapping_metrics: DefaultDict[MetricKey, JobExecutionMetric] = defaultdict(lambda: JobExecutionMetric(
        count=0,
        max_duration=0,
        min_duration=None,
        sum_duration=0,
    ))
    mapping_max_epoch_ends: DefaultDict[MetricKey, int] = defaultdict(lambda: 0)

    latest_epoch_start = retrieve_latest_epoch_start()

    # retrieve all metrics which may still have steps to be updated and populate our mapping
    metrics: List[JobExecutionMetric] = riberry.model.conn.query(JobExecutionMetric).filter_by(
        epoch_start=latest_epoch_start
    ).all()

    for metric in metrics:
        key = MetricKey(
            epoch_start=metric.epoch_start,
            epoch_end=metric.epoch_end,
            job_execution_id=metric.job_execution_id,
            stream_name=metric.stream_name,
            step_name=metric.step_name,
        )
        mapping_metrics[key] = metric

    query_steps = latest_query_steps(epoch_start=latest_epoch_start, limit=limit)
    for step_data in query_steps.yield_per(10_000):
        step_name, step_dt_start, step_dt_completed, stream_name, job_execution_id, form_id = step_data

        # Convert the step's start and completed date-times to timestamps
        epoch_start = pendulum.UTC.convert(pendulum.instance(step_dt_start)).timestamp()
        epoch_end = pendulum.UTC.convert(pendulum.instance(step_dt_completed)).timestamp()

        # Find the duration of the step + its epoch range, rounded up/down with the given interval
        range_start = round_down(epoch_end, time_interval)
        range_end = round_down(epoch_end + time_interval, time_interval)

        key = MetricKey(
            epoch_start=range_start,
            epoch_end=range_end,
            job_execution_id=job_execution_id,
            stream_name=stream_name,
            step_name=step_name,
        )
        metric = mapping_metrics[key]

        # Avoid re-processing the same step by comparing the last epoch value to populate the metric
        if metric.epoch_last and epoch_end <= metric.epoch_last:
            continue

        # Update the max epoch end for each metric
        mapping_max_epoch_ends[key] = max(mapping_max_epoch_ends[key], epoch_end)

        # Populate the metric's attributes
        metric.epoch_start = range_start
        metric.epoch_end = range_end
        metric.form_id = form_id
        metric.job_execution_id = job_execution_id
        metric.stream_name = stream_name
        metric.step_name = step_name
        metric.count += 1
        duration = epoch_end - epoch_start
        metric.sum_duration += duration

        if duration > metric.max_duration:
            metric.max_duration = max(metric.max_duration, duration)
        if metric.min_duration is None or duration < metric.min_duration:
            metric.min_duration = min(metric.min_duration or duration, duration)

    # Collect all update metrics and updated the "updated" timestamp
    updated_timestamp = pendulum.DateTime.utcnow()
    for key, max_epoch in mapping_max_epoch_ends.items():
        metric = mapping_metrics[key]
        metric.epoch_last = max_epoch
        metric.updated = updated_timestamp
        updated.append(metric)

    return updated


def process_metrics(time_interval: int, limit: int):
    """ Updates all metrics with the latest steps for the given interval. """

    updated = updated_metrics(time_interval=time_interval, limit=limit)
    if updated:
        logger.info(f'Updated {len(updated)} metric instances')
        riberry.model.conn.bulk_save_objects(updated)
        riberry.model.conn.commit()



import shutil

import click

import riberry
from ..root import cli


@cli.command('conf', help='Copy the given environment config to Riberry\'s default path')
@click.argument('path', required=True)
def import_config(path):
    target = str(riberry.config.CONF_DEFAULT_PATH)
    shutil.copy(path, target)
    print(f'Copied {path!r} to {target!r}.')



import threading
from contextlib import contextmanager
from queue import Empty

import time

import riberry
from ..exc import Defer
from ..task_queue import TaskQueue, Task

log = riberry.log.make(__name__)


def execution_listener(task_queue: TaskQueue):
    try:
        task: Task = task_queue.queue.get(timeout=2.0)
    except Empty:
        return

    execution_thread = threading.Thread(
        name=f'{task.execution_id}:{task.id}:{task.definition.name}',
        target=execute,
        args=(task, task_queue)
    )
    execution_thread.start()


def execute(task: Task, task_queue: TaskQueue):
    try:

        job_execution: riberry.model.job.JobExecution = riberry.model.conn.query(
            riberry.model.job.JobExecution
        ).filter_by(
            id=task.execution_id,
        ).one()

        if job_execution.task_id == task.id:
            task_scope = execute_entry_task(job_execution=job_execution, task=task)
        else:
            task_scope = execute_receiver_task(job_execution=job_execution, task=task)

        context_scope = riberry.app.current_context.scope(
            root_id=job_execution.task_id,
            task_id=task.id,
            task_name=task.definition.name,
            stream=task.definition.stream,
            category=None,
            step=task.definition.step,
        )

        with context_scope, task_scope:
            execute_task(job_execution=job_execution, task=task)

    finally:
        task_queue.counter.decrement()


@contextmanager
def execute_entry_task(job_execution: riberry.model.job.JobExecution, task: Task):
    riberry.app.actions.executions.execution_started(
        task_id=task.id,
        root_id=job_execution.task_id,
        job_id=task.execution_id,
        primary_stream=task.definition.stream,
    )

    yield


# noinspection PyUnusedLocal
@contextmanager
def execute_receiver_task(job_execution: riberry.model.job.JobExecution, task: Task):
    yield


def execute_task(job_execution: riberry.model.job.JobExecution, task: Task):
    riberry.app.util.task_transitions.task_active(
        context=riberry.app.current_context,
        props={},
    )
    riberry.app.current_riberry_app.backend.set_active_task(task=task)

    status = 'SUCCESS'
    start_time = time.time()
    try:
        log.info('Starting task %s', task.definition.name)
        task.definition.func()

    except Defer:
        status = None

    except Exception as exc:
        status = 'FAILURE'
        riberry.app.actions.artifacts.create_artifact_from_traceback(category='Fatal')
        log.exception('Failed with exception: %s', exc)

    finally:
        end_time = time.time()
        log.info('Completed task %s in %.4f seconds', task.definition.name, end_time - start_time)

        riberry.app.util.task_transitions.task_complete(
            context=riberry.app.current_context,
            props={},
            state=status or 'IGNORED',
        )

        if status:
            riberry.app.actions.executions.execution_complete(
                task_id=task.id,
                root_id=job_execution.task_id,
                status=status,
                stream=task.definition.stream,
                context=riberry.app.current_context,
            )

        riberry.app.current_riberry_app.backend.set_active_task(task=None)



import logging

import riberry
from ..task_queue import TaskQueue

log = logging.getLogger(__name__)


def background(queue: TaskQueue):
    riberry.app.tasks.echo()
    with queue.lock:
        if not queue.limit_reached():
            riberry.app.tasks.poll(track_executions=True, filter_func=lambda _: not queue.limit_reached())
        else:
            log.debug('Queue limit reached, skipped task polling')

    riberry.app.tasks.refresh()



import logging
import os
import sys
import threading

import riberry


class Filter(logging.Filter):

    def filter(self, record):
        if riberry.app.current_context.current.root_id:
            record.root_id = f'root={riberry.app.current_context.current.root_id}'
        else:
            record.root_id = '-'

        if riberry.app.current_context.current.task_id:
            record.prefix = f'[id={riberry.app.current_context.current.task_id}] '
        else:
            record.prefix = ''

        if riberry.app.current_context.current.step:
            record.context = f'{riberry.app.current_context.current.step}'
        elif threading.current_thread() == threading.main_thread():
            record.context = 'backend.main'
        else:
            record.context = record.threadName

        return record


def configure(log_level='ERROR'):
    handler = riberry.log.stdout_handler
    handler.formatter = logging.Formatter(
        os.environ.get(
            'RIBERRY_LOGFORMAT',
            '%(levelname)-8s | %(asctime)-s | %(context)-41s | %(root_id)-41s | %(prefix)s%(message)s'
        )
    )

    for filter_ in list(handler.filters):
        handler.removeFilter(filter=filter_)

    handler.addFilter(Filter())
    riberry.log.logger.addHandler(handler)
    riberry.log.logger.setLevel(log_level.upper())



import logging
import os
import sys

# noinspection PyTypeChecker
root_name: str = None

# noinspection PyTypeChecker
logger: logging.Logger = None
stdout_handler = logging.StreamHandler(stream=sys.stdout)


def make(name: str) -> logging.Logger:
    return logging.getLogger(name=name)


def init():
    logger.addHandler(logging.NullHandler())
    stdout_handler.formatter = logging.Formatter(
        os.environ.get('RIBERRY_LOGFORMAT', '%(levelname)-8s %(asctime)-15s %(name)s - %(message)s')
    )
    if os.environ.get('RIBERRY_LOGLEVEL'):
        logger.setLevel(level=os.environ['RIBERRY_LOGLEVEL'].upper())
        logger.addHandler(stdout_handler)
        logger.info('Logging to stdout...')



from typing import List

import riberry

log = riberry.log.make(__name__)


def _tracker_key(value):
    return f'workflow:active:{value}'


class CeleryExecutionTracker(riberry.app.backends.tracker.RiberryExecutionTracker):

    def _check_stale_executions(
            self,
            executions: List[riberry.model.job.JobExecution],
            app_instance: riberry.model.application.ApplicationInstance
    ):
        redis = riberry.celery.util.celery_redis_instance()
        for execution in executions:
            if not redis.sismember(_tracker_key(app_instance.internal_name), execution.task_id):
                self._cancel_execution(execution=execution)

    def track_execution(self, root_id: str, app_instance: riberry.model.application.ApplicationInstance):
        redis = riberry.celery.util.celery_redis_instance()
        instance = riberry.app.env.get_instance_name()
        key = _tracker_key(instance)
        log.debug(f'Tracking execution: root={root_id!r}, key={key!r}')
        redis.sadd(key, root_id)

    def _artifact_message(self, execution: riberry.model.job.JobExecution):
        return (
            f'The current executions\'s ID ({execution.task_id}) was not found within Redis and has '
            f'therefore been cancelled. This usually occurs when Redis is flushed while an execution is '
            f'in the READY or ACTIVE state.'
        )
